#! /usr/bin/perl -w

# Lang.pm: language specific code
# 
#    Copyright (C) 2005, Patrice Dumas
# 
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
use strict;

package CMZ::Lang;

use File::Basename;
use Date::Format;

my $default_ext = '';
our $cmfile_ext = '.cm';
our $catalog_ext = '.cmcat';

my %special_strings = (
	'dateqq' => \&do_dateqq,
	'timeqq' => \&do_timeqq,
	'versqq' => \&do_versqq,
	'qftitle' => \&do_qftitlech,
	'qftitlch' => \&do_qftitlech,
	'authqq' => \&do_authqq,
	'vidqq' => \&do_vidqq,
);

my @lt = localtime(time);

my %lang = (
	'c' => {'suffix' => {'deck'=>'c', 'keep'=>'h'},
	'include' => '#include "%f"',
	'comment' => '/*%f*/',
	'comment_regexp' => '^\\s*\\/\\*',
	'dateqq' => '#define IDATQQ=%f ',
	'timeqq' => '#define ITIMQQ=%f',
	'versqq' => '#define VERSQQ "%f"',
        'iversq' => '#define IVERSQ %f',
	'authqq' => '#define AUTHQQ "%f"',
	'vidqq' => "static char vidqq[] = \"@(#)%f\";",
	},
	'f77' => {'suffix' => {'deck'=>'F', 'keep'=>'inc'},
	'include' => '#include "%f"',
	'comment_char' => '*',
	'dateqq' => '      IDATEQQ=%f',
	'timeqq' => '      ITIMEQQ=%f',
	'versqq' => "      VERSQQ='%f'",
	'iversq' => '      IVERSQ=%f',
	'qftitle' => '     + %fH',
	'qftitlch' => '     +',
	'authqq' => "       AUTHQQ='%f'",
	},
	'ftn' => {'suffix' => {'deck'=>'f', 'keep'=>'inc'},
	#'include' => '      INCLUDE \'%f\'',
	'comment_char' => '*'},
	'mtn' => {'suffix' => {'deck'=>'mtn', 'keep'=>'mti'},
	'include' => '&u#base/%f',
	# in cmz comments * is used and not !???
	'comment_char' => '!', 'cmz_comment_char' => '*'},
	'text' => {'suffix' => {'deck'=>'txt', 'keep'=>'txt'},
	'no_comment' => 1},
	'' => {'suffix' => {'keep'=>''}},
	'kumac' => {'suffix' => {'keep'=>'kumac', 'deck'=>'kumac'},
	'comment_char' => '*'},
	'cdf' => {'suffix' => {'keep'=>'txt', 'deck'=>'txt'},
	},
#	'include' => '%f'},
#	'' => {'suffix' => {'deck'=>'', 'keep'=>''},
#	'include' => '%f'},
    );

my %common_strings_lang = ( 'f77' => ['ftn'] );
my %alias_lang = ( 'c' => ['xcc','cc'] );

my %alias_strings_lang = ();
my %alias_lang_lang = ();

foreach my $base_lang(keys(%common_strings_lang))
{
	$alias_strings_lang{$base_lang} = $base_lang;
	foreach my $alias(@{$common_strings_lang{$base_lang}})
	{
		$alias_strings_lang{$alias} = $base_lang;
		foreach my $string (keys(%special_strings))
		{
			$lang{$alias}->{$string} = $lang{$base_lang}->{$string}
				if (exists($lang{$base_lang}->{$string}));
		}
	}
}

foreach my $base_lang (keys(%lang))
{
	$alias_lang_lang{$base_lang} = $base_lang;
	$alias_strings_lang{$base_lang} = $base_lang unless 
		(exists($alias_strings_lang{$base_lang}));
	if (exists($alias_lang{$base_lang}))
	{
		foreach my $alias(@{$alias_lang{$base_lang}})
		{
			$lang{$alias} = $lang{$base_lang};
			$alias_lang_lang{$alias} = $base_lang;
		}
	}
}

my %suffix_to_lang = ();

foreach my $lang (keys(%lang))
{
	next if ($lang eq '');
	if (exists($lang{$lang}->{'suffix'}))
	{
		foreach my $element(('deck', 'keep'))
		{
			if (exists($lang{$lang}->{'suffix'}->{$element}))
			{
				my $suffix_entry = $lang{$lang}->{'suffix'}->{$element};
				if (exists($suffix_to_lang{$suffix_entry}) and
					$suffix_to_lang{$suffix_entry}->{'lang'} eq $lang)
				{# same suffix for keep and decks
					$suffix_to_lang{$suffix_entry}->{'directive'} = undef;
					delete $suffix_to_lang{$suffix_entry}->{'directive'};
				}
				elsif (!exists($suffix_to_lang{$suffix_entry}))
				{
					$suffix_to_lang{$suffix_entry} = 
						{'lang' => $lang, 'directive' => $element};
				}
			}
		}
	}
}

my $full_title;
my ($title,$version,$date);
$title = 'Unknown';
$version = '0.00/00';
$date = '01/01/1990';

sub get_title($)
{
	$full_title = shift;
	my $string = $full_title;
	if ($string =~ s/^(\w[^\s]*\s*)//)
	{
		$title = $1;
		if ($string =~ s/^([\d\.\/]+\s*)//)
		{
			$version = $1;
			if ($string =~ s/^([\d\/]+\s*)//)
			{
				$date = $1;
			}
		}
	}
}

my %patch_langs = ('$kumacs' => 'kumac',
	'$version' => '');

sub patch_lang($)
{
	my $patch_name = shift;
	return $patch_langs{lc($patch_name)} 
		if (exists($patch_langs{lc($patch_name)}));
	return undef;
}
	

sub suffix_to_directive($)
{
	my $suffix = shift;
	return $suffix_to_lang{$suffix} if (exists($suffix_to_lang{$suffix}));
	return undef;
}

sub filename($$$)
{
	my $directive = shift;
	my $lang = shift;
	my $name = shift;
	my $suffix = $default_ext;
	$lang = lc($lang);
#	print STDERR "filename: $directive, $lang, $name\n";
	$suffix = ".".$lang{$lang}->{'suffix'}->{lc($directive)} if 
	(exists($lang{$lang}) and exists($lang{$lang}->{'suffix'}) and
		exists($lang{$lang}->{'suffix'}->{lc($directive)}));
	return (lc($name).$suffix);
}

sub cmfile($$$)
{
	my $directive = shift;
	my $lang = shift;
	my $name = shift;
	my $suffix = '';
	return basename(lc($name)) if (CMZ::Directives::is_seq(lc($directive)));
	$lang = lc($lang);
	$suffix = $lang{$lang}->{'suffix'}->{lc($directive)} if 
	(exists($lang{$lang}) and exists($lang{$lang}->{'suffix'}) and
		exists($lang{$lang}->{'suffix'}->{lc($directive)}));
	return (lc($name).$cmfile_ext.$suffix);
}

sub include($)
{
	my $lang = shift;
	return ($lang{lc($lang)}->{'include'}) if (exists($lang{lc($lang)}->{'include'}));
	return undef;
}

sub comment_char($)
{
	my $lang = shift;
	return ($lang{lc($lang)}->{'comment_char'}) if (exists($lang{lc($lang)}->{'comment_char'}));
	return undef;
}

sub cmz_comment_char($)
{
	my $lang = shift;
	return ($lang{lc($lang)}->{'cmz_comment_char'}) if (exists($lang{lc($lang)}->{'cmz_comment_char'}));
	return comment_char($lang);
}
	
sub comment_text($)
{
	my $lang = shift;
	return ($lang{lc($lang)}->{'comment'}) if (exists($lang{lc($lang)}->{'comment'}));
	return undef;
}

sub differing_string_lang($$)
{
	my $lang1 = shift;
	my $lang2 = shift;
	return 0 if ($alias_strings_lang{lc($lang1)} eq $alias_strings_lang{lc($lang2)});
	return 1;
}

sub differing_lang($$)
{
	my $lang1 = shift;
	my $lang2 = shift;
	return 0 if ($alias_lang_lang{lc($lang1)} eq $alias_lang_lang{lc($lang2)});
	return 1;
#	return 0 if (lc($lang1) eq lc($lang2));
#	return 0 if (exists($alias_lang{$lang1}) 
#		and (grep {lc($lang2) eq lc($_)} @{$alias_lang{lc($lang1)}}));
#	return 0 if (exists($alias_lang{lc($lang2)}) 
#		and (grep {lc($lang1) eq lc($_)} @{$alias_lang{lc($lang2)}}));
#	foreach my $aliased_lang(keys %alias_lang)
#	{
		#print STDERR "--> $lang1 $lang2 $aliased_lang: @{$alias_lang{$aliased_lang}} \n";
#		if ((grep {lc($lang1) eq lc($_)} @{$alias_lang{$aliased_lang}})
#			and (grep {lc($lang2) eq lc($_)} @{$alias_lang{$aliased_lang}}))
#		{
#			return 0;
#		}
#	}
#	return 1;
}

# substitute %f with $string and %% with % in $text.
sub substitute_string($$$)
{
	my $string = shift;
	my $text = shift;
	my $verbose = shift;
	$text = '' if (!defined($text));
	$string = '' if (!defined($string));
	my $result = '';
	while ($text)
	{
		#print STDERR "write_include: $text\n";
		if ($text =~ /%/)
		{
			#print STDERR "write_include: process %\n";
			$text =~ s/^([^%]*)%//;
			$result .= $1;
			if ($text =~ s/^%//)
			{
				$result .= '%';
			}
			elsif ($text =~ s/^f//)
			{
				$result .= $string;
			}
			elsif ($text =~ s/^(.)//)
			{
				$result .= $1;
			}
		}
		else
		{
			$result .= $text;
			$text = '';
		}
	}
	print STDERR "Resulting text: $result\n" if ($verbose > 1);
#	print STDERR "Empty include for $lang, $seq_name\n" if ($result = '' and $verbose > 1);
	return $result;
}

sub include_line($$$)
{
	my $lang = shift;
	my $filename = shift;
	my $verbose = shift;
	my $text = include($lang);
	$text = '' if (!defined($text));
	return substitute_string($filename,$text, $verbose);
}

sub comment_line($$$;$)
{
	my $lang = shift;
	$lang = lc($lang);
	my $comment = shift;
	my $verbose = shift;
	# true if the comment is in cmz preamble. In that case the rules 
	# are different
	my $in_cmz_comment = shift;
	die "Bug: comment_line comment undef\n" if (!defined($comment));
	die "Bug: comment_line lang unset\n" if (!defined($lang) or $lang =~/^\s*$/);
	return $comment if ($lang{$lang}->{'no_comment'});
	my $regexp = $lang{$lang}->{'comment_regexp'};
	#print STDERR "$regexp\n" if defined($regexp);
	return $comment if ((defined($regexp)) and ($comment =~ /$regexp/));
	chomp($comment);
	my $comment_char;
	if ($in_cmz_comment)
	{
		$comment_char = cmz_comment_char($lang);
	}
	else
	{
		$comment_char = comment_char($lang);
	}
	if (defined($comment_char))
	{ 
		my $quoted_char = quotemeta($comment_char);
		return $comment."\n" if ($comment =~ /^$quoted_char/);
		return $comment_char.$comment."\n";
	}
	my $comment_text = comment_text($lang);
	if (defined($comment_text))
	{
		return (substitute_string($comment, $comment_text, $verbose) ."\n");
	}
	print STDERR "No comment format for $lang\n" if ($verbose > 1);
	return undef;
}

sub special($$$$)
{
	my $lang = shift;
	$lang = lc($lang) if (defined($lang));
	my $special = shift;
	my $special_arg = shift;
	my $verbose = shift;
	return &{$special_strings{$special}}($special,$special_arg,$lang,$verbose)
		if (exists($special_strings{$special}));
	return '';
}

sub do_dateqq ($$$$)
{
	my $special = shift;
	my $special_arg = shift;
	my $lang = shift;
	my $verbose = shift;
	return (substitute_string(time2str('%Y%m%d',time), $lang{$lang}->{$special}, $verbose) ."\n") 
		if (exists($lang{$lang}->{$special}));
	return '';
}

sub do_timeqq ($$$$)
{
	my $special = shift;
	my $special_arg = shift;
	my $lang = shift;
	my $verbose = shift;
	return (substitute_string(time2str('%H%M',time),$lang{$lang}->{$special}, $verbose) ."\n") 
		if (exists($lang{$lang}->{$special}));
	return '';
}

sub do_versqq ($$$$)
{
	my $special = shift;
	my $special_arg = shift;
	my $lang = shift;
	my $verbose = shift;
	my $vers = '';
	$vers .= substitute_string('00.00/00', $lang{$lang}->{$special},$verbose) ."\n"
		if (exists($lang{$lang}->{$special}));
	$vers .= substitute_string('000000',$lang{$lang}->{'iversq'}, $verbose) ."\n"
		if (exists($lang{$lang}->{'iversq'}));
	return $vers;
}

sub do_authqq ($$$$)
{
	my $special = shift;
	my $special_arg = shift;
	my $lang = shift;
	my $verbose = shift;
	return (substitute_string('somebody', $lang{$lang}->{$special},$verbose) ."\n") 
		if (exists($lang{$lang}->{$special}));
	return '';
}

sub do_vidqq ($$$$)
{
	my $special = shift;
	my $special_arg = shift;
	my $lang = shift;
	my $verbose = shift;
	return (substitute_string($title.$version.$date,$lang{$lang}->{$special}, $verbose) ."\n") 
		if (exists($lang{$lang}->{$special}));
	return '';
}

sub do_qftitlech ($$$$)
{
	my $special = shift;
	my $special_arg = shift;
	my $lang = shift;
	my $verbose = shift;
	if (!differing_lang($lang,'f77') or !differing_string_lang($lang,'f77'))
	{
		$special_arg = 8 if (!defined($special_arg));
		my $trunc_title = substr($full_title,0,$special_arg);
		$trunc_title .= " " x ($special_arg -length($trunc_title));
		return (substitute_string($special_arg, $lang{'f77'}->{$special},$verbose) . $trunc_title . "\n");
	}
	else
	{
		return (substitute_string($full_title, $lang{$lang}->{$special},$verbose) ."\n") 
		if (exists($lang{$lang}->{$special}));
	}
	return '';
}

1;
